import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.Handlers";
import "moebius92.Misc";
import "moebius92.RaceTrack";

local quickKeys = { 0x1000002A, 0x10000032, 0x1000003E, 0x10000047, 0x10000050, 0x10000057,
                    0x1000005F, 0x10000067, 0x10000071, 0x1000004F, 0x10000056, 0x1000005E,

                    0x10000066, 0x10000070, 0x10000078, 0x10000082, 0x1000008E, 0x10000096,
                    0x1000009E, 0x10000011, 0x10000019, 0x10000022, 0x10000029, 0x10000031,

                    0x1000003D, 0x10000046, 0x1000004E, 0x10000055, 0x1000005D, 0x10000005,
                    0x1000000B, 0x10000010, 0x10000018, 0x10000021, 0x10000028, 0x10000030,

                    0x1000003C, 0x10000045, 0x1000004D, 0x1000008D, 0x10000095, 0x1000009D,
                    0x100000A3, 0x100000AD, 0x100000B0, 0x100000B9, 0x10000004, 0x1000000A,

                    0x1000000F, 0x10000020, 0x10000027, 0x1000002F, 0x1000003B, 0x10000044,
                    0x1000004C, 0x10000054, 0x1000005C, 0x10000065, 0x1000006F, 0x100000AC,

                    0x100000AF, 0x100000B8, 0x10000003, 0x10000009, 0x1000000E, 0x10000017,
                    0x1000001F, 0x10000026, 0x1000002E, 0x1000009C, 0x100000A2, 0x100000AB };

local keymap = { };
for k, v in pairs(quickKeys) do
	keymap[v] = k;
end

skills = {
	["Bane Flare"]					= { name = "bane",		recovery =  true, shortcut = "0x7000B60D", duration = 15, act = -0.172, start = 2.6, stop = 1.5, r = 1, g = 1, b = 0, effect = "Dazed" },
	["Blinding Flash"]				= { name = "blinding",	recovery =  true, shortcut = "0x70003152", duration = 30, act =  0.000, start = 1.4, stop = 0.3, r = 1, g = 1, b = 0, effect = "Dazed" },
	["Cracked Earth"]				= { name = "cracked",	recovery = false, shortcut = "0x70003143", duration = 30, act = -1.250, start = 7.3, stop = 0.0, r = 0, g = 0, b = 1, effect = "The Shaking of Arda" },
	["Ents go to War"]				= { name = "march",		recovery =  true, shortcut = "0x70000FA7", duration =  6, act = -1.547, start = 0.0, stop = 0.0, r = 1, g = 0, b = 0, effect = "Stunned" },
	["Herb-lore"]					= { name = "herb",		recovery = false, shortcut = "0x70003EC0", duration = 30, act = -1.250, start = 0.0, stop = 0.0, r = 0, g = 0, b = 1, effect = "Rooted" },
	["Light of the Rising Dawn"]	= { name = "light",		recovery =  true, shortcut = "0x70003EC2", duration =  3, act = -1.188, start = 1.4, stop = 0.3, r = 1, g = 0, b = 0, effect = "Stunned" },
	["Storm-lore"]					= { name = "storm",		recovery =  true, shortcut = "0x7000B63C", duration =  3, act = -1.234, start = 0.0, stop = 0.0, r = 1, g = 0, b = 0, effect = "Stunned" },
	["Test of Will"]				= { name = "test",		recovery =  true, shortcut = "0x70003EB9", duration =  5, act = -0.641, start = 3.0, stop = 1.9, r = 1, g = 0, b = 0, effect = "Stunned" }
};

ccQS = {
	["Dazed"]						= "0x70003152",
	["Feared"]						= "0x70003E82",
	["Recovering"]					= 0,
	["Rooted"]						= "0x70003EC0",
	["Stunned"]						= "0x70003EC2",
	["Vulnerable"]					= "0x70003F0D"
};

function NewCCElement(size)
	local element = { };

	element.control = Turbine.UI.Control();
	element.quickslot = Turbine.UI.Lotro.Quickslot();
	element.control:SetSize(size, size);
	element.control:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	element.control:SetZOrder(1);
	element.control:SetVisible(false);
	element.quickslot:SetParent(element.control);
	element.quickslot:SetPosition(-1, -1);
	element.quickslot:SetSize(size, size);
	element.quickslot:SetMouseVisible(false);
	return element;
end

CCTimerRegime = class(Turbine.UI.Control);

function CCTimerRegime:Constructor(msg)
	Turbine.UI.Control.Constructor(self);
	self.msg = msg;
	self:SetMouseVisible(false);
end

function CCTimerRegime:MouseEnter(args)
	Turbine.Shell.WriteLine(self.msg);
end

function CCTimerRegime:MouseLeave(args)
	Turbine.Shell.WriteLine("---");
end

CCTimerWindow = class(Turbine.UI.Window);

function CCTimerWindow:Constructor(version)
	Turbine.UI.Window.Constructor(self);

	self.x = 0;
	self.y = 0;
	self.preroot = 10;
	self.version = version;

	for _, v in pairs(skills) do
		v.useTime = 0;
		v.hitTime = 0;
	end

	self:LoadSettings();

	self.hide = false;
	self.visible = false;
	self.size = 36;
	self.seconds = 30;
	self.dragging = false;
	self.move = false;
	self:SetPosition(self.x, self.y);
	self:SetMouseVisible(false);
	self.track = moebius92.RaceTrack.RaceTrack(self.size, 0);
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.name = self.player:GetName();
	self.tskills = self.player:GetTrainedSkills();
	self.duration = { };
	self.cc = { };

	self.fnEffectAdded = function(sender, args)
		local e = self.targetEffects:Get(args.Index);
		local name = e:GetName();

		if self.time ~= nil then
			for k, v in pairs(skills) do
				if v.effect == name then
					self.time.cc = e:GetStartTime();
					self:Calibrate();
					break;
				end
			end
		end

		if ccQS[name] ~= nil then
			self.targetCCName = name;
			self.targetCC.endTime = e:GetStartTime() + e:GetDuration();
			self:UpdateLayout();
			self:Update();
			self:SetWantsUpdates(true);
		end
	end;
	self.fnEffectRemoved = function(sender, args)
		if args.Effect:GetID() == self.targetCCID then
			self.targetCCName = nil;
			self.targetCCID = nil;
			self:UpdateLayout();
			self:Update();
			self:SetWantsUpdates(true);
		end
	end;
	self.fnEffectsCleared = function(sender, args)
		if self.targetCCName ~= nil then
			self.targetCCName = nil;
			self.targetCCID = nil;
			self:UpdateLayout();
			self:Update();
			self:SetWantsUpdates(true);
		end
	end;

	self.targetCC = NewCCElement(self.size);
	self.targetCC.control:SetParent(self);
	self.targetCC.startTime = nil;

	self:HookSkills();
	moebius92.Handlers.AddChatHandler(function(m) self:ChatMessage(m, Turbine.Engine.GetGameTime()) end, Turbine.ChatType.PlayerCombat, self.name);
	moebius92.Handlers.AddEventHandler(self.tskills, "SkillAdded", function() self:HookSkills(); end);
	moebius92.Handlers.AddEventHandler(self.tskills, "SkillRemoved", function() self:HookSkills(); end);
	moebius92.Handlers.AddEventHandler(self.player, "TargetChanged", function() self:StopCalibrate(); self:UpdateTarget(); end);

	self:SetBackColor(Turbine.UI.Color(0.25, 0.0, 1.0, 0.0));
	self:SetWantsKeyEvents(true);
	self:Layout();
	self:SetVisible();
	self:UpdateTarget();
end

function CCTimerWindow:UpdateTarget()
	if self.target ~= nil then
		for k, v in pairs(self.targetEffectsHandlers) do
			moebius92.Handlers.RemoveEventHandler(self.targetEffects, k, v);
		end
	end

	self.target = nil;
	self.targetEffects = nil;
	self.targetEffectsHandlers = nil;
	self.targetCCName = nil;
	self.targetCCID = nil;

	self.target = self.player:GetTarget();
	if self.target ~= nil and self.target:IsA(Turbine.Gameplay.Actor) then
		self.targetEffects = self.target:GetEffects();
		self.targetEffectsHandlers = { };
		self.targetEffectsHandlers.EffectAdded = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectAdded", self.fnEffectAdded);
		self.targetEffectsHandlers.EffectRemoved = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectRemoved", self.fnEffectRemoved);
		self.targetEffectsHandlers.EffectsCleared = moebius92.Handlers.AddEventHandler(self.targetEffects, "EffectsCleared", self.fnEffectsCleared);

		for i = 1, self.targetEffects:GetCount(), 1 do
			local e = self.targetEffects:Get(i);
			local name = e:GetName();

			if ccQS[name] ~= nil then
				self.targetCC.endTime = e:GetStartTime() + e:GetDuration();
				self.targetCCName = name;
				self.targetCCID = e:GetID();
				break;
			end
		end
	else
		self.target = nil;
	end

	self:UpdateLayout();
	self:Update();
	self:SetWantsUpdates(true);
end

function CCTimerWindow:HookSkills()
	for _, v in pairs(skills) do
		if v.activeSkill ~= nil then
			for k1, v1 in pairs(v.handlers) do
				moebius92.Handlers.RemoveEventHandler(v.activeSkill, k1, v1);
			end
		end

		v.handlers = nil;
		v.activeSkill = nil;
	end

	for i = 1, self.tskills:GetCount(), 1 do
		local askill = self.tskills:GetItem(i);
		local name = askill:GetSkillInfo():GetName();

		if skills[name] ~= nil then
			skills[name].activeSkill = askill;
			skills[name].handlers = { };
			skills[name].handlers["ResetTimeChanged"] = moebius92.Handlers.AddEventHandler(askill, "ResetTimeChanged", function() self:SkillUse(name); end);
		end
	end
end

function CCTimerWindow:SkillUse(skill)
	local time = Turbine.Engine.GetGameTime();
	local cooldown = skills[skill].activeSkill:GetCooldown();
	local rtime = skills[skill].activeSkill:GetResetTime();

	if rtime >= 0 and (math.abs(time - (rtime - cooldown)) < 0.1) then
		skills[skill].useTime = rtime - cooldown;
		if (skills[skill].useTime - skills[skill].hitTime) < 1.5 then
			if self.time ~= nil then
				self.time.reset = skills[skill].useTime;
				self:Calibrate();
			end
			self:AddCC(skill);
		end
	end
end

function CCTimerWindow:ChatMessage(m, time)
	local i;
	local name, rest;
	local skill, type;

	m = string.gsub(m, "<[^>]*>", "");
	i = string.find(m, " ");

	if i ~= nil then
		name = trim(string.sub(m, 1, i - 1));
		rest = trim(string.sub(m, i + 1));
		if name == self.name then
			if string.sub(rest, 1, 9) == "scored a " then
				local j = string.find(rest, " with ");
				i = string.find(rest, " on ");
				if i ~= nil then
					skill = string.sub(rest, j + 6, i - 1);
					if string.sub(skill, 1, 9) == "Improved " then
						skill = string.sub(skill, 10);
					end
					if string.sub(skill, 1, 9) == "Advanced " then
						skill = string.sub(skill, 10);
					end

					if skills[skill] ~= nil then
						skills[skill].hitTime = time;
						if (skills[skill].hitTime - skills[skill].useTime) < 1.5 then
							if self.time ~= nil then
								self.time.reset = skills[skill].useTime;
								self:Calibrate();
							end
							self:AddCC(skill);
						end
					end
				end
			end
		end
	end
end

function CCTimerWindow:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end

	Turbine.UI.Control.SetVisible(self, self.visible and not self.hide);
end

function CCTimerWindow:Layout()
	local w = self.size + 4;
	local h = self.size + self.seconds * 10 + 10;

	if self.regimes ~= nil then
		for k, v in pairs(self.regimes) do
			v:SetParent(nil);
		end
	end

	self.regimes = { };

	self:SetSize(w, h);

	self.regimes.recovery = CCTimerRegime("recovery state\non entry, it is safe to use fan the flames on stunned targets");
	self.regimes.recovery:SetParent(self);
	self.regimes.recovery:SetBackColor(Turbine.UI.Color(0.25, 1.0, 0.0, 0.0));
	self.regimes.recovery:SetSize(w, 10);
	self.regimes.recovery:SetPosition(0, h - 10);
	self.regimes.recovery.start = 0.0;
	self.regimes.recovery.stop = -1.0;
	self.regimes.recovery.color = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);

	local s = { };
	local e = { };
	local t = { };

	for k, v in pairs(skills) do
		if v.start ~= v.stop then
			s[k] = { };
			s[k].start = v.start * 10;
			if v.effect == "Rooted" or v.effect == "The Shaking of Arda" then
				local m = 0;

				for k1, v1 in pairs(skills) do
					if v1.start > m and v1.start < v.start and v.start ~= v.stop then
						m = v1.start;
					end
					s[k].stop = m * 10;
				end
			else
				s[k].stop = v.stop * 10;
			end
			s[k].r = v.r;
			s[k].g = v.g;
			s[k].b = v.b;
		end
	end

	for k, v in pairs(s) do
		e[v.start] = { n = 0, r = 0, g = 0, b = 0 };
		e[v.stop] = { n = 0, r = 0, g = 0, b = 0 };
	end

	for k, v in pairs(s) do
		e[v.start].n = e[v.start].n + 1;
		e[v.start].r = e[v.start].r + v.r;
		e[v.start].g = e[v.start].g + v.g;
		e[v.start].b = e[v.start].b + v.b;

		e[v.stop].n = e[v.stop].n - 1;
		e[v.stop].r = e[v.stop].r - v.r;
		e[v.stop].g = e[v.stop].g - v.g;
		e[v.stop].b = e[v.stop].b - v.b;
	end

	for k, v in pairs(e) do
		table.insert(t, k);
	end
	table.sort(t, function (a, b) return a > b end);

	local r = 0;
	local g = 0;
	local b = 0;
	local n = 0;
	local rc, gc, bc, mc;

	for k, v in ipairs(t) do
		n = n + e[v].n;
		r = r + e[v].r;
		g = g + e[v].g;
		b = b + e[v].b;

		if n > 0 then
			msg = "";
			for k1, v1 in pairs(s) do
				if v == v1.start then
					if skills[k1].effect == "Dazed" or skills[k1].effect == "Stunned" then
						msg = msg .. "on entry, it is not safe to use " .. k1 .. " on Dazed targets\n";
					else
						msg = msg .. "on entry, it is safe to use " .. k1 .. " on Rooted targets\n";
					end
				end
				if t[k + 1] == v1.stop then
					if skills[k1].effect == "Dazed" then
						msg = msg .. "on exit, it is safe to use " .. k1 .. " on Dazed/Stunned targets\n";
					elseif skills[k1].effect == "Stunned" then
						msg = msg .. "on exit, it is safe to use " .. k1 .. " on Dazed targets\n";
					end
				end
			end

			self.regimes[v] = CCTimerRegime(msg);
			self.regimes[v]:SetParent(self);
			rc = r / n;
			gc = g / n;
			bc = b / n;
			mc = math.max(rc, gc, bc);
			if n > 1 then
				rc = 1.0 - rc / mc;
				gc = 1.0 - gc / mc;
				bc = 1.0 - bc / mc;
				mc = math.max(rc, gc, bc);
			end
			if mc > 0 then
				self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, rc / mc, gc / mc, bc / mc));
				self.regimes[v].color = Turbine.UI.Color(1.0, rc / mc, gc / mc, bc / mc);
			else
				self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, 0.0, 0.0, 0.0));
				self.regimes[v].color = Turbine.UI.Color(1.0, 0.0, 0.0, 0.0);
			end
			self.regimes[v]:SetMouseVisible(false);
			self.regimes[v]:SetSize(w, v - t[k + 1]);
			self.regimes[v]:SetPosition(0, h - v - 10);
			self.regimes[v].start = v / 10;
			self.regimes[v].stop = t[k + 1] / 10;
		end
	end

	self:UpdateLayout();
end

function CCTimerWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.move then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function CCTimerWindow:MouseMove(args)
	if self.dragging then
		local x, y = self:GetPosition();
		x = x + (args.X - self.dragStartX);
		y = y + (args.Y - self.dragStartY);
		self.x = x;
		self.y = y;
		self:SetPosition(x, y);
	end
end

function CCTimerWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self:SaveSettings();
		self.dragging = false;
	end
end

function CCTimerWindow:AssignLanes()
	local time = Turbine.Engine.GetGameTime();

	self.track:set(1);
	for _, v in ipairs(self.cc) do
		if v.startTime > time then
			v.lane = nil;
		else
			local y = math.floor((v.endTime - time) * 10);
			v.lane = self.track:place(-y);
		end
	end

	self:UpdateLayout();
end

function CCTimerWindow:UpdateLayout()
	local w = 0;

	if self.targetCCName == nil then
		self.targetCC.control:SetVisible(false);
		for k, v in pairs(self.cc) do
			v.vis = true;
			if v.lane ~= nil and v.lane > w then
				w = v.lane
			end
		end
	else
		local found = false;

		self.targetCC.control:SetVisible(true);
		for k, v in pairs(self.cc) do
			if not found and (self.targetCCName == "Recovering" and skills[v.type].recovery == true and math.abs(self.targetCC.endTime - (v.endTime + 1.0)) < 0.1) or
			                 (skills[v.type].effect == self.targetCCName and math.abs(self.targetCC.endTime - v.endTime) < 0.1) then
				self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, skills[v.type].shortcut));
				v.vis = false;
				found = true;
			else
				v.vis = true;
				if v.lane ~= nil and v.lane > w then
					w = v.lane
				end
			end
		end

		if not found then
			if type(ccQS[self.targetCCName]) == "string" then
				self.targetCC.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, ccQS[self.targetCCName]));
			end
		end
	end

	w = self.size * (w + 1) + 4;
	self:SetWidth(w);
	for k, v in pairs(self.regimes) do
		v:SetWidth(w);
	end
	self:SetVisible(false);
	self:SetVisible(true);
end

function CCTimerWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local delta;
	local changed = false;
	local t = { };

	for k, v in ipairs(self.cc) do
		delta = v.endTime - time;
		if delta < 0.0 and v.pre then
			changed = true;
			v.pre = false;
			v.endTime = v.endTime + skills[v.type].duration;
			v.type = "Herb-lore";
			v.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, skills["Herb-lore"].shortcut));
		elseif delta < -1.0 or (delta < 0.0 and not skills[v.type].recovery) then
			changed = true;
			v.control:SetParent(nil);
			v.control:SetVisible(false);
			table.insert(t, 1, k);
		elseif v.startTime <= time and v.lane == nil then
			changed = true;
		end
	end

	for _, v in ipairs(t) do
		table.remove(self.cc, v);
	end

	if changed then
		table.sort(self.cc, function(a, b) return a.endTime < b.endTime; end);
		self:AssignLanes();
	end

	local t;

	for _, v in ipairs(self.cc) do
		if v.lane == nil then
			v.control:SetVisible(false);
		else
			t = v.endTime - time;

			v.control:SetVisible(v.vis);
			v.control:SetPosition(self.size * v.lane + 2, (self.seconds - t) * 10);
			v.control:SetBackColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
			for k1, v1 in pairs(self.regimes) do
				if t <= v1.start and t >= v1.stop then
					v.control:SetBackColor(v1.color);
					break;
				end
			end
		end
	end

	if self.targetCCName ~= nil then
		t = self.targetCC.endTime - time;
		if t < 0 then
			self.targetCCName = nil;
			self:UpdateLayout();
		end
		if self.targetCCName == "Recovering" then
			t = t - 1.0;
		end
		self.targetCC.control:SetPosition(2, (self.seconds - t) * 10);
		self.targetCC.control:SetBackColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
		for k1, v1 in pairs(self.regimes) do
			if t <= v1.start and t >= v1.stop then
				self.targetCC.control:SetBackColor(v1.color);
				break;
			end
		end
	end

	if #self.cc == 0 and self.targetCCName == nil then
		self:SetWantsUpdates(false);
		if not self.move then
			self:SetVisible(false);
		end
	end
end

function CCTimerWindow:AddCC(skill)
	local item = { };

	item = NewCCElement(self.size);
	item.control:SetParent(self);
	item.type = skill;
	item.startTime = skills[skill].useTime + skills[skill].act;
	item.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, skills[skill].shortcut));
	if skill == "Cracked Earth" then
		item.endTime = item.startTime + self.preroot;
		item.pre = true;
	else
		item.endTime = item.startTime + skills[skill].duration;
		item.pre = false;
	end

	table.insert(self.cc, item);
	table.sort(self.cc, function(a, b) return a.endTime < b.endTime; end);

	self:AssignLanes();
	self:Update();
	self:SetVisible(true);
	self:SetWantsUpdates(true);

	skills[skill].useTime = 0;
	skills[skill].hitTime = 0;
end

function CCTimerWindow:KeyDown(args)
	local time = Turbine.Engine.GetGameTime();

	if args.Action == 0x1000007B then
		if self.move then
			self.move = false;
			if #self.cc == 0 and self.targetCCName == nil then
				self:SetVisible(false);
			end
			if self.dragging then
				self:SaveSettings();
				self.dragging = false;
			end
		else
			self.move = true;
			self:SetVisible(true);
		end

		self:SetMouseVisible(self.move);
		for k, v in pairs(self.regimes) do
			v:SetMouseVisible(self.move);
		end
	elseif args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	elseif keymap[args.Action] ~= nil and self.time ~= nil then
		self.time.key = time;
	end
end

function CCTimerWindow:StartCalibrate()
	self.time = { };
end

function CCTimerWindow:Calibrate()
	if self.time.cc ~= nil and self.time.reset ~= nil then
		Turbine.Shell.WriteLine(string.format("Act: %.3f", self.time.cc - self.time.reset));
		if self.time.key ~= nil then
			Turbine.Shell.WriteLine(string.format("Key: %.3f", self.time.cc - self.time.key));
		end
		self.time = nil;
	end
end

function CCTimerWindow:StopCalibrate()
	self.time = nil;
end

function CCTimerWindow:SaveSettings()
	local t = { };

	t.version = self.version;

	t.x = self.x;
	t.y = self.y;

	t.act = { };
	t.start = { };
	t.stop = { };
	t.r = { };
	t.g = { };
	t.b = { };
	for k, v in pairs(skills) do
		t.act[k] = v.act;
		t.start[k] = v.start;
		t.stop[k] = v.stop;
		t.r[k] = v.r;
		t.g[k] = v.g;
		t.b[k] = v.b;
	end

	t.durations = { };
	t.durations["Bane Flare"] = skills["Bane Flare"].duration;
	t.durations["Blinding Flash"] = skills["Blinding Flash"].duration;
	t.preroot = self.preroot;

	Turbine.PluginData.Save(Turbine.DataScope.Character, "CCTimer", serialize(t));
end

function CCTimerWindow:LoadSettings()
	local t = Turbine.PluginData.Load(Turbine.DataScope.Character, "CCTimer");
	if type(t) == "string" then
		t = loadstring("return " .. t .. ";")();
	end

	if type(t) == "table" then
		local v = snarfversion(t.version);

		if v[1] == 0 and v[2] == 4 then
			self.x = t.x;
			self.y = t.y;

			for k, v in pairs(t.act) do
				skills[k].act = v;
			end

			for k, v in pairs(t.start) do
				skills[k].start = v;
			end

			for k, v in pairs(t.stop) do
				skills[k].stop = v;
			end

			for k, v in pairs(t.r) do
				skills[k].r = v;
			end

			for k, v in pairs(t.g) do
				skills[k].g = v;
			end

			for k, v in pairs(t.b) do
				skills[k].b = v;
			end

			for k, v in pairs(t.durations) do
				skills[k].duration = v;
			end

			self.preroot = t.preroot;
		end
	end
end

CCTimerCommand = class(Turbine.ShellCommand);

function CCTimerCommand:Constructor(window)
	Turbine.ShellCommand.Constructor(self);

	self.window = window;
end

function CCTimerCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	if args[1] == "calibrate" then
		self.window:StartCalibrate();
	elseif args[1] == "preroot" then
		if tonumber(args[2]) ~= nil then
			self.window.preroot = tonumber(args[2]);
			self.window:Layout();
			self.window:SaveSettings();
		end

		Turbine.Shell.WriteLine("Cracked Earth pre-root time is " .. self.window.preroot);
	elseif args[1] == "safe" then
		local skill = nil;

		for k, v in pairs(skills) do
			if v.name == args[2] then
				skill = k;
			end
		end

		if skill == nil then
			for k, v in pairs(skills) do
				if v.start == v.stop then
					Turbine.Shell.WriteLine(v.name .. ": " .. k .. " (safe) is disabled.");
				else
					Turbine.Shell.WriteLine(v.name .. ": " .. k .. " (safe) is " .. v.start .. " - " .. v.stop .. "(" .. v.r .. "," .. v.g .. "," .. v.b ..")");
				end
			end
		else
			if args[3] == "disable" then
				skills[skill].start = 0;
				skills[skill].stop = 0;
				self.window:Layout();
				self.window:SaveSettings();
			elseif tonumber(args[3]) ~= nil and tonumber(args[4]) ~= nil then
				skills[skill].start = tonumber(args[3]);
				skills[skill].stop = tonumber(args[4]);
				if tonumber(args[5]) ~= nil and tonumber(args[6]) ~= nil and tonumber(args[7]) ~= nil then
					skills[skill].r = tonumber(args[5]);
					skills[skill].g = tonumber(args[6]);
					skills[skill].b = tonumber(args[7]);
				end
				self.window:Layout();
				self.window:SaveSettings();
			end

			if skills[skill].start == skills[skill].stop then
				Turbine.Shell.WriteLine(skill .. ": " .. k .. " (safe) is disabled.");
			else
				Turbine.Shell.WriteLine(skill .. ": " .. k .. " (safe) is " .. skills[skill].start .. " - " .. skills[skill].stop .. "(" .. skills[skill].r .. "," .. skills[skill].g .. "," .. skills[skill].b ..")");
			end
		end
	elseif args[1] == "act" then
		local skill = nil;

		for k, v in pairs(skills) do
			if v.name == args[2] then
				skill = k;
			end
		end

		if skill == nil then
			for k, v in pairs(skills) do
				Turbine.Shell.WriteLine(v.name .. ": " .. k .. " (act) is " .. v.act);
			end
		else
			if tonumber(args[3]) ~= nil then
				skills[skill].act = tonumber(args[3]);
				self.window:SaveSettings();
			end

			Turbine.Shell.WriteLine(skill .. " (act) is " .. skills[skill].act);
		end
	elseif args[1] == "duration" then
		local skill = nil;

		if args[2] == "blinding" then
			skill = "Blinding Flash";
		elseif args[2] == "bane" then
			skill = "Bane Flare";
		end

		if skill == nil then
			Turbine.Shell.WriteLine("bane: Bane Flare (duration) is " .. skills["Bane Flare"].duration);
			Turbine.Shell.WriteLine("blinding: Blinding Flash (duration) is " .. skills["Blinding Flash"].duration);
		else
			if tonumber(args[3]) ~= nil then
				skills[skill].duration = tonumber(args[3]);
				self.window:SaveSettings();
			end

			Turbine.Shell.WriteLine(skill .. " (duration) is " .. skills[skill].duration);
		end
	else
		Turbine.Shell.WriteLine("usage: /cctimer <act <cc> <#>|duration <cc> <#>|<safe> <cc> <start> <stop> [r] [g] [b]|preroot <#>|calibrate>");
	end
end

function CCTimerCommand:GetShortHelp()
	return "Loremaster Crowd Control Timer";
end

plugin.command = CCTimerCommand(CCTimerWindow(plugin:GetVersion()));
Turbine.Shell.AddCommand("cctimer", plugin.command);
Turbine.Shell.AddCommand("cc", plugin.command);